//coroutine 2311, see stl2.c

#include <optional>
#include <experimental/generator> //needed

//coroutine1 231119 co_yield based on experimental::generator
std::experimental::generator<int> cofunc_generator()
{
	for (int i = 0; i < 3; ++i)
	{
		co_yield i;
	}

};
void Test_generator()
{
	cout << "Test_generator: call by loop" << endl;
	for (int i : cofunc_generator())
		cout << i << " ";
	cout << endl;
	cout << "Test_generator: call by resume" << endl;
	std::experimental::generator<int> hReturnGen = cofunc_generator();
	std::experimental::generator<int>::iterator it = hReturnGen.begin();
	int iFirst = *it;
	assert(*it == 0);
	assert(*it._Coro.promise()._Value == 0);
	it++;
	assert(*it == 1);
	assert(*it._Coro.promise()._Value == 1);
	it._Coro.resume();
	assert(*it._Coro.promise()._Value == 2);
	assert(*it == 2);
	cout << endl;
}

//coroutine2
template<std::movable T>
class CReturnObjGen
{
public:
	struct promise_type
	{
		std::optional<T> mValue;
		CReturnObjGen<T> get_return_object()
		{
			return CReturnObjGen{ Handle::from_promise(*this) };
		}
		static suspend_always initial_suspend() noexcept
		{
			return {};
		}
		static suspend_always final_suspend() noexcept
		{
			return {};
		}
		suspend_always yield_value(T value) noexcept
		{
			mValue = std::move(value);
			return {};
		}
		void await_transform() = delete;// Disallow co_await in generator coroutines.
		[[noreturn]] static void unhandled_exception() { throw; }
		void return_void() noexcept {} //wxc compiler error without this https://stackoverflow.com/questions/67306337/a-coroutines-promise-must-declare-either-return-value-or-return-void-erro
	};
	using Handle = coroutine_handle<promise_type>;
	explicit CReturnObjGen(const Handle coroutine) : mhCoroutine{ coroutine }
	{}
	CReturnObjGen() = default;
	~CReturnObjGen()
	{
		if (mhCoroutine)
			mhCoroutine.destroy();
	}
	CReturnObjGen(const CReturnObjGen&) = delete;
	CReturnObjGen& operator=(const CReturnObjGen&) = delete;
	CReturnObjGen& operator=(CReturnObjGen&& other) noexcept
	{
		if (this != &other)
		{
			if (mhCoroutine)
				mhCoroutine.destroy();
			mhCoroutine = other.mhCoroutine;
			other.mhCoroutine = {};
		}
		return *this;
	}
	Handle getHandle() { return mhCoroutine; } //wxc
private:
	Handle mhCoroutine;
};
template<std::integral T>
CReturnObjGen<T> cofuncYield(T begin, const T end)
{
	while (begin < end)
	{
		cout << "cofuncYield: before co_yield, will yield " << begin << endl;
		co_yield begin++;
		cout << "cofuncYield: after co_yield" << endl;
	}
}
void coroutineYieldTest()
{
	cout << "coroutineYieldTest started" << endl;
#ifdef WXC_SINGLE_CALL
	auto c = cofunc_generator();
	cout << " single call to cofunc_generator() returned " << c;
	int r = c.get();
	r = c();
#endif
	//cout << "testing cofunc_generator:" << endl;
	//for (int i : cofunc_generator()) //good for (auto& i : cofunc_generator())  //error: int& i , int i is good.
	//	cout << " " << i;
	cout << endl;
	cout << "testing CReturnObjGen<T> cofuncYield:" << endl; //ouput  0 1 2
	CReturnObjGen<int> RangeReturn = cofuncYield(68, 70); //auto
	CReturnObjGen<int>::Handle hRange = RangeReturn.getHandle();
	CReturnObjGen<int>::promise_type& promise = hRange.promise();
	hRange(); //output before co_yield, first = 68
	assert(promise.mValue == 68);
	hRange.resume(); //output after co_yield before co_yield, first = 69
	assert(promise.mValue == 69);
	hRange.resume(); //output after co_yield
	bool bDone = hRange.done();
	if (!bDone)
		hRange(); //crash after reaching end
	cout << "coroutineYieldTest ended" << endl;
}

//coroutine3 https://riowingwp.wordpress.com/2023/12/10/coroutine-3
//#define WXC_CALLER_KEEP_COPY_OF_HANDLE //both works
//#define WXC_CROSS_CHECK_HANDLE //still not working, trying to let WXC_CALLER_KEEP_COPY_OF_HANDLE routine return valid returnobj
//#define WXC_BY_RESUME //both works
struct CReturnObj 
{
	struct promise_type
	{
		unsigned mValue; //let both has it
		CReturnObj get_return_object()
		{ //this name required
#ifdef WXC_CALLER_KEEP_COPY_OF_HANDLE
#ifdef WXC_CROSS_CHECK_HANDLE
	//CReturnObj ro = { mHandle = coroutine_handle<promise_type>::from_promise(*this) }
	//ro.mHandle..promise()
	//cout << "get_return_object" << endl;
	//return ro; //.mHandle = std wxc
			return { .mHandleDebug = coroutine_handle<promise_type>::from_promise(*this) //a CReturnObj inited by handle
		#else
			return {};
		#endif
#else
			return {.mHandle = coroutine_handle<promise_type>::from_promise(*this) //a CReturnObj inited by handle
		};
#endif
			}
		suspend_never initial_suspend() { return {}; }
		suspend_never final_suspend() noexcept { return {}; }
		void unhandled_exception() {}
		void return_void() noexcept {} //wxc
		};
#ifdef WXC_CALLER_KEEP_COPY_OF_HANDLE
#ifdef WXC_CROSS_CHECK_HANDLE
	coroutine_handle<promise_type> mHandleDebug;
	operator coroutine_handle<promise_type>() const { return mHandleDebug; }
	operator coroutine_handle<>() const { return mHandleDebug; }
#else
#endif
#else
	coroutine_handle<promise_type> mHandle;
	operator coroutine_handle<promise_type>() const { return mHandle; }
	// A coroutine_handle<promise_type> converts to coroutine_handle<>
	operator coroutine_handle<>() const { return mHandle; }
#endif
	};
#ifdef WXC_CALLER_KEEP_COPY_OF_HANDLE //both good 231102
bool CALLER_KEEP_COPY_OF_HANDLE = true;
struct CAwaiterSetHandleInCaller
{
	coroutine_handle<CReturnObj::promise_type>* mpPodToSetHandleInCaller;
	constexpr bool await_ready() const noexcept
	{
		return false; // optimization, if true, no suspent on co_await
	}
	void await_suspend(coroutine_handle<CReturnObj::promise_type> h)
	{
		if (!mpPodToSetHandleInCaller->address())
			*mpPodToSetHandleInCaller = h; //only first call is null and that's all we need.
	}
	constexpr void await_resume() const noexcept {}  //could be  co_await expression
};
CReturnObj cofuncAwait(coroutine_handle<CReturnObj::promise_type>* pPodToSetHandleInCaller)
#else
bool CALLER_KEEP_COPY_OF_HANDLE = false;
template<typename PromiseType> struct CAwaiterPromise //for the first co_await, just to recv handle
{
#ifdef WXC_CROSS_CHECK_HANDLE
	//coroutine_handle<>* mpToSetHandleInCaller;
#endif
	PromiseType* mPromise;
	bool await_ready() { return false; } // says yes call await_suspend
	bool await_suspend(coroutine_handle<PromiseType> h)
	{
		mPromise = &h.promise();
#ifdef WXC_CROSS_CHECK_HANDLE
		//if (!mpToSetHandleInCaller->address())
		//	*mpToSetHandleInCaller = h;
#endif
		return false;
	}
	PromiseType* await_resume() { return mPromise; }
};
//co_await 231102
//#ifdef WXC_CROSS_CHECK_HANDLE
//	CReturnObj cofuncAwait(coroutine_handle<>* pPodToSetHandleInCaller) //WXC_CO_RETURN was ReturnObject3
//#else
CReturnObj cofuncAwait() //WXC_CO_RETURN was ReturnObject3
//#endif
#endif
//caller start -> callee enter -> after CAwaiterSetHandleInCaller -> loop enter before co_await -> caller call 
//output pattern: caller loop start ->   i=, after -> loop enter before co_await ->  caller loop end.
//recap: 
{
	cout << "\tcounter: entered" << endl;
	bool bFirstEntry = true;
#ifdef WXC_CALLER_KEEP_COPY_OF_HANDLE
	CAwaiterSetHandleInCaller waiter{ pPodToSetHandleInCaller }; //pPodToSetHandleInCaller will recv value thru  await_suspend and eventually crh in main makes sense
#else
	//#ifdef WXC_CROSS_CHECK_HANDLE
	//	auto myPromiseInRoutine = co_await CAwaiterPromise<CReturnObj::promise_type>{pPodToSetHandleInCaller}; //was only for WXC_CO_RETURN, to get myPromiseInRoutine
	//#else
	auto myPromiseInRoutine = co_await CAwaiterPromise<CReturnObj::promise_type>{}; //was only for WXC_CO_RETURN, to get myPromiseInRoutine
	//#endif
#endif
	cout << "\tcounter: after CAwaiterSetHandleInCaller, inside loop below: " << endl;
	for (unsigned i = 0;; ++i)
	{
		cout << "\t\tcounter: loop enter, before co_await | ";
#ifdef WXC_CALLER_KEEP_COPY_OF_HANDLE
		CReturnObj::promise_type& myPromiseInRoutine = pPodToSetHandleInCaller->promise();
		if (!bFirstEntry)
			myPromiseInRoutine.mValue = i; //&myPromiseInRoutine not null and not valid on first entry
		co_await waiter;
#else
		myPromiseInRoutine->mValue = i; //WXC_CO_RETURN
		co_await suspend_always{};
#endif
		cout << "cofuncAwait: i = " << i << ", after co_await," << "bFirstEntry=" << bFirstEntry << ", loop end | " << endl;
		bFirstEntry = false;
	}
}

void coroutineWaitTest()
{
	cout << "coroutineWaitTest: start, CALLER_KEEP_COPY_OF_HANDLE = " << CALLER_KEEP_COPY_OF_HANDLE << endl;
#ifdef WXC_CALLER_KEEP_COPY_OF_HANDLE
	coroutine_handle<CReturnObj::myPromise_type> crh, hDebug; //no sense until it somehow go thru routine see cpp.t
	//#ifdef WXC_CROSS_CHECK_HANDLE
	//	hDebug = cofuncAwait(&crh);
	//#else
	cofuncAwait(&crh);
	//#endif
	CReturnObj::promise_type& myPromiseInCaller = crh.promise();
#else
	coroutine_handle<CReturnObj::promise_type> crh = cofuncAwait(); //promise_type for WXC_CO_RETURN
	CReturnObj::promise_type& myPromiseInCaller = crh.promise();
#endif
	cout << "coroutineWaitTest: cofuncAwait called" << endl;
	for (int i = 0; i < 3; ++i)
	{
		cout << "coroutineWaitTest: loop start" << endl;
		std::cout << "	return : " << myPromiseInCaller.mValue << std::endl;
#ifdef WXC_CALLER_KEEP_COPY_OF_HANDLE
		CReturnObj::promise_type& promiseDebug = hDebug.promise();
#endif
#ifdef WXC_BY_RESUME
		crh.resume();
#else
		crh();
#endif
		cout << "coroutineWaitTest: loop end" << endl;
	}
	crh.destroy();
}